module irq(rst, clk, irq, in_exc, rti, in_irq, irq_stage1, irq_stage2, irq_stage3, irq_done, entering_irq);
   input  rst, clk;
   input  irq;		// interrupt
   reg	  irq_pending;	// save irq (in case exception)
   input  in_exc;	// if in exception handler, irq should not be handled
   input  rti;		// RTI, for counting irq depth
   output in_irq;	// executing irq handler
   reg [3:0] irq_depth; // recursive irq depth, at most 15
   output irq_stage1;	// the previous cycle has irq
   reg    irq_stage1;
   output irq_stage2;	// the previous previous cycle has irq
   reg    irq_stage2;
   output irq_stage3;
   reg    irq_stage3;
   output irq_done;	// irq is finished
   reg	  irq_done;
   output entering_irq;
   assign entering_irq = irq_stage1 | irq_stage2 | irq_stage3;

   // irq: input signal (async)
   always@(posedge clk or negedge rst)
     begin
       if (~rst)
         begin
         irq_stage1 <= 0;
	 irq_stage2 <= 0;
	 irq_stage3 <= 0;
	 irq_done <= 0;
	 end
       // if handling exception, irq is not allowed
       else if (~in_exc)
	 begin
         irq_stage1 <= irq | irq_pending;
	 irq_stage2 <= irq_stage1;
	 irq_stage3 <= irq_stage2;
	 irq_done <= irq_stage3;
	 end
     end

   // save irq if handling exception
   always@(posedge clk or negedge rst)
     begin
       if (~rst)
         irq_pending <= 0;
       else if (in_exc & irq)
         irq_pending <= 1;
       else if (~in_exc) // irq is handled
         irq_pending <= 0;
     end
   
   assign in_irq = (irq_depth != 0);
   
   always@(posedge clk or negedge rst)
     begin
        if (~rst)
	  irq_depth <= 0;
	else if (in_exc)
	  ;
	else if (irq | irq_pending)
	  irq_depth <= irq_depth + 1;
	else if (rti)
	  irq_depth <= irq_depth - 1;
     end
endmodule // irq
