module stall(rst, clk, flush, bubble, indirect, indirected, load_mem, arith_alu, indirect_mem, ra1, ra2, wa1, wa2);
   input  rst, clk;
   input  flush;	// should flush stall regs when flush
   output bubble;	// pause PC, ALU, register, clear memory
   output indirect; 	// pause PC, pipeline, register and memory write
   output indirected; 	// the previous stage was indirected
   
   // the characterics of the current status of components
   input  load_mem;	// memory is doing a read
   input  arith_alu;	// ALU is doing arithmetic
   input  indirect_mem; // the instruction in memory needs indirect stage

   input [2:0] ra1, ra2;// read address in EXEC stage
   input [2:0] wa1; 	// write address in EXEC stage
   input [2:0] wa2; 	// write address in MEM stage

   reg 	  indirected;	// the previous cycle was indirected
   reg 	  bubbled;	// the previous cycle was bubbled

   /* We want `stall' to behave this way:
    *   After a instruction has been detected to "should stall" after some combintorial logic after clock posedge,
    *       the stall signal should go up before clock negedge.
    *   If the clock is negative, the stall signal should not change.
    *   A cycle after this instruction is indirected,
    *       the indirect signal should go down (after the posedge of next cycle).
    *   A cycle after this instruction is bubbled, if the indirect signal was down in last cycle,
    *       the bubble signal should go down (after the posedge of next cycle).
    */
   assign indirect = ~indirected && indirect_mem;
   // (hazard unsolvable by forwarding) LD R0,PCoffset; ADD R1,R1,R0;
   // (hazard solvable by reading regs at MEM stage, but we did not do so) LD R0,PCoffset; ST R0,PCoffset;
   assign bubble = (~bubbled || indirected) && load_mem && (ra1 == wa2 || ra2 == wa2);

   // We do not want to stall more than one cycle.
   // When the pipeline is stalled, pipeline registers are frozen, and how can we know when to recover the pipeline?
   always@(posedge clk or negedge rst)
     if (~rst)
       begin
	  bubbled <= 0;
	  indirected <= 0;
       end
     else if(flush)
       begin
	  bubbled <= 0;
	  indirected <= 0;
       end
     else
       begin
	  bubbled <= bubble;
	  indirected <= indirect;
       end

endmodule // stall
